#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest
import daggen as dg
import networkx as nx
import igraph as ig

class TestDAG(unittest.TestCase):

    def test_dag_creation(self):
        d = dg.DAG(seed=42, num_tasks=10)
        dot_str = "digraph G {\n" \
                  "  1 [computation=68719476736, alpha=0.06]\n" \
                  "  2 [computation=8589934592, alpha=0.06]\n" \
                  "  2 -> 3 [data=33554432]\n" \
                  "  2 -> 4 [data=33554432]\n" \
                  "  3 [computation=24450842215, alpha=0.02]\n" \
                  "  3 -> 5 [data=75497472]\n" \
                  "  3 -> 7 [data=75497472]\n" \
                  "  4 [computation=231928233984, alpha=0.20]\n" \
                  "  4 -> 6 [data=301989888]\n" \
                  "  5 [computation=28991029248, alpha=0.11]\n" \
                  "  5 -> 8 [data=75497472]\n" \
                  "  5 -> 9 [data=75497472]\n" \
                  "  6 [computation=370921150755, alpha=0.16]\n" \
                  "  7 [computation=9613001727, alpha=0.18]\n" \
                  "  7 -> 8 [data=411041792]\n" \
                  "  8 [computation=18552871712, alpha=0.04]\n" \
                  "  8 -> 10 [data=838860800]\n" \
                  "  9 [computation=1128959608590, alpha=0.18]\n" \
                  "  10 [computation=123125765851, alpha=0.10]\n" \
                  "}"
        self.assertEqual(str(d), dot_str)

    def test_seed(self):
        seed = 42
        d1 = dg.DAG(seed=seed, num_tasks=15)
        t1, e1 = d1.task_n_edge_dicts()
        d2 = dg.DAG(seed=seed, num_tasks=15)
        t2, e2 = d2.task_n_edge_dicts()
        self.assertEqual(e2, e2)
        self.assertEqual(t1, t2)
        seed = 43
        d3 = dg.DAG(seed=seed, num_tasks=15)
        t3, e3 = d3.task_n_edge_dicts()
        self.assertNotEqual(t1, t3)
        self.assertNotEqual(e1, e3)

    def test_tasks_and_edges(self):
        d = dg.DAG(seed=42, num_tasks=15)
        tasks, edges = d.task_n_edge_dicts()
        self.assertEqual(len(tasks), 15)
        self.assertEqual(len(edges), 13)

    def test_igraph(self):
        d = dg.DAG(seed=42, num_tasks=15)
        tasks, edges = d.task_n_edge_dicts()
        g = ig.Graph.DictList(vertices=tasks, edges=edges, directed=True)
        igraph_out = "IGRAPH DN-- 15 13 --\n" \
                     "+ attr: alpha (v), computation (v), name (v), data (e), source (e), target (e)\n" \
                     "+ edges (vertex names):\n" \
                     "1->4, 2->3, 3->5, 3->6, 4->7, 6->9, 7->8, 9->10, 9->11, 10->13, 11->12,\n" \
                     "12->15, 13->14"
        self.assertEqual(str(g), igraph_out)

    def test_networkx(self):
        d = dg.DAG(seed=46, num_tasks=15)
        tasks, edges = d.task_n_edge_tuples()
        net_dag = nx.DiGraph()
        net_dag.add_nodes_from(tasks)
        net_dag.add_edges_from(edges)
        task_degrees = [(1, 1), (2, 1), (3, 2), (4, 2), (5, 4), (6, 1),
                        (7, 3), (8, 3), (9, 2), (10, 1), (11, 3),
                        (12, 4), (13, 2), (14, 2), (15, 1)]
        self.assertEqual(task_degrees, list(net_dag.degree))

    def test_ccr(self):
        d = dg.DAG(seed=46, num_tasks=15, ccr=3)


if __name__ == '__main__':
    unittest.main()



# distutils: language = c

from libc.stdio cimport FILE

# cdef class CCRType:
#     cdef public int MIXED=0, N_2, N_LOG_N, N_3
#
#     def __init__(self, mixed, n_2, n_log_n, n_3):
#         self.MIXED = mixed
#         self.N_2 = n_2
#         self.N_LOG_N = n_log_n
#         self.N_3 = n_3

cdef extern from "daggen_commons.h":
    ctypedef struct Global:
        int n
        double fat
        double regular
        int ccr
        double density
        double mindata, maxdata
        double minalpha, maxalpha
        int jump
        FILE* output_file

    Global config
    ctypedef (_Task *) Task
    ctypedef (_DAG *) DAG

    cdef enum complexity_t:
        MIXED=0,
        N_2,
        N_LOG_N, # (n2 log(n2) indeed
        N_3

    # ctypedef class CCRType [object complexity_t]:
    #     cdef:
    #         int MIXED "MIXED"
    #         int N_2 "N_2"
    #         int N_LOG_N "N_LOG_N"
    #         int N_3 "N_3"

    cdef struct _Task:
        int tag
        double cost
        double data_size
        double alpha
        int nb_children
        Task * children
        double * comm_costs
        int * transfer_tags
        complexity_t complexity

    cdef struct _DAG:
        int nb_levels
        int * nb_tasks_per_level
        Task ** levels

    void outputDAG(DAG dag)
    void outputDOT(DAG dag)

ctypedef Global* ConfigPointer



# distutils: language = c
#
from . cimport common

__all__ = [
    "DAG",
    # "CCRType"
]

cdef extern from "daggen_commons.c":
    void outputDAG(common.DAG dag)
    void outputDOT(common.DAG dag)

cdef extern from "daggen.c":
    common.DAG generateDAG()
    void freeDAG(common.DAG dag)

cdef class DAG:
    cdef common.DAG _dag
    cdef common.ConfigPointer _conf
    cdef int _seed

    cdef void _generate_DAG(self)
    cdef void _release_DAG(self)
    cdef _init_params(self, int, double, double, int, double,
                      long, long, double, double, int)





#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""A setuptools based setup module.

See:
https://packaging.python.org/guides/distributing-packages-using-setuptools/
https://github.com/pypa/sampleproject
"""

from setuptools import setup, Extension
try:
    from Cython.Build import cythonize
except (NameError, ModuleNotFoundError):
    def cythonize(*args, **kwargs):
        pass

extensions = [
        Extension('daggen.common', ['daggen/common.pyx']),
        Extension('daggen.dag', ['daggen/dag.pyx'], extra_compile_args=[])
    ]

setup(
    packages=['daggen'],
    name='daggen',
    ext_modules=cythonize(extensions, language_level=3)
)




# distutils: language = c
#
from . cimport common
from libc.stdio cimport stdout
from typing import Tuple

# Default DAG values obtained from the C implementation
DEF NUM_TASKS = 100
DEF FAT = 0.5
DEF REGULAR = 0.9
DEF DENSITY = 0.5
DEF MIN_DATA = 2048
DEF MAX_DATA =  11264
DEF MIN_ALPHA = 0.0
DEF MAX_ALPHA = 0.2
DEF JUMP_SIZE = 1
DEF CCR = 0

DEF TASK_SIZE_LABEL = "computation"
DEF DATA_COMM_LABEL = "data"

cdef extern from "stdlib.h":
    void srand(int seed)

cdef extern from "daggen.c":
    common.DAG generateDAG()
    void freeDAG(common.DAG dag)
    common.Global config

cdef _validate_0_to_1_arg(double value, str name):
    if value < 0 or value > 1:
        raise ValueError(
            f"Unsupported {name} value {value:.2f}"
        )

cdef _validate_smaller_than(double smaller_v, double larger_v,
                            str name_smaller, str name_larger):
    if smaller_v > larger_v:
        raise ValueError(
            f"Value for {name_larger} ({larger_v}) must "
            f"be greater than {name_smaller} ({smaller_v})"
        )

cdef _validate_positive_arg(double value, str name):
    if value < 0:
        raise ValueError(
            f"Unsupported {name} value {value:.2f}"
        )

cdef _validate_ccr(int ccr):
    if ccr < 0 or ccr > 3:
        raise ValueError(
            f"Unsupported ccr value {ccr}"
        )

cdef class DAG:

    def __init__(self, seed: int,
                 num_tasks: int = NUM_TASKS,
                 min_data: int = MIN_DATA,
                 max_data: int = MAX_DATA,
                 min_alpha: float = MIN_ALPHA,
                 max_alpha: float = MAX_ALPHA,
                 fat: float = FAT,
                 density: float = DENSITY,
                 regular: float = REGULAR,
                 ccr: int = CCR,
                 jump_size: int = JUMP_SIZE
                 ):
        self._seed = seed
        srand(self._seed)
        self._init_params(num_tasks, fat, regular, ccr,
                          density, min_data, max_data,
                          min_alpha, max_alpha, jump_size)
        self._generate_DAG()

    def __cinit__(self):
        self._conf = &config
        self._dag = NULL
        self._init_params(NUM_TASKS, FAT, REGULAR, CCR,
                          DENSITY, MIN_DATA, MAX_DATA,
                          MIN_ALPHA, MAX_ALPHA, JUMP_SIZE)


    def __dealloc__(self):
        self._release_DAG()

    cdef void _generate_DAG(self):
        self._dag = generateDAG()

        # Tag all the nodes
        cdef int i, j
        cdef node_count = 1     # Tags start at 1

        for i in range(self._dag.nb_levels):
            for j in range(self._dag.nb_tasks_per_level[i]):
                self._dag.levels[i][j].tag = node_count
                node_count += 1

    cdef void _release_DAG(self):
        if self._dag is not NULL:
            freeDAG(self._dag)

    cdef _init_params(self, int num_tasks, double fat,
                      double regular, int ccr, double density,
                      long min_data, long max_data,
                      double min_alpha, double max_alpha, int jump_size):
        # Validate all argument values
        _validate_positive_arg(num_tasks, <str>"num_tasks")
        _validate_positive_arg(min_data, <str> "min_data")
        _validate_positive_arg(max_data, <str> "max_data")
        _validate_0_to_1_arg(min_alpha, <str> "min_alpha")
        _validate_0_to_1_arg(max_alpha, <str> "max_alpha")
        _validate_smaller_than(min_data, max_data, <str>"min_data", <str>"max_data")
        _validate_smaller_than(min_alpha, max_alpha, <str> "min_alpha", <str> "max_alpha")
        _validate_0_to_1_arg(fat, <str>"fat")
        _validate_0_to_1_arg(density, <str> "density")
        _validate_ccr(ccr)

        self._conf.n = num_tasks
        self._conf.fat = fat
        self._conf.regular = regular
        self._conf.ccr = ccr
        self._conf.density = density
        self._conf.mindata = min_data
        self._conf.maxdata = max_data
        self._conf.minalpha = min_alpha
        self._conf.maxalpha = max_alpha
        self._conf.jump = jump_size
        self._conf.output_file = stdout

    def __str__(self):
        result = "digraph G {\n"
        for i in range(self._dag.nb_levels):
            # obtain the vertices
            for j in range(self._dag.nb_tasks_per_level[i]):
                task = self._dag.levels[i][j]
                result += f"  {task.tag} [{TASK_SIZE_LABEL}={task.cost:.0f}, alpha={task.alpha:.2f}]\n"

                # Get the edges
                for k in range(task.nb_children):
                    result += f"  {task.tag} -> {task.children[k].tag} [{DATA_COMM_LABEL}={task.comm_costs[k]:.0f}]\n"

        result += "}"
        return result

    def task_n_edge_dicts(self) -> Tuple[list[dict], list[dict]]:
        tasks = []
        edges = []
        for i in range(self._dag.nb_levels):
            # obtain the vertices
            for j in range(self._dag.nb_tasks_per_level[i]):
                task = self._dag.levels[i][j]
                tasks.append({'name': task.tag,
                              TASK_SIZE_LABEL: int(task.cost),
                              'alpha': round(task.alpha, 2)})

                # Get the edges
                for k in range(task.nb_children):
                    edges.append({'source' : task.tag,
                                  'target': task.children[k].tag,
                                  DATA_COMM_LABEL: int(task.comm_costs[k])})

        return tasks, edges

    def task_n_edge_tuples(self) -> Tuple[list[Tuple], list[Tuple]]:
        tasks = []
        edges = []
        for i in range(self._dag.nb_levels):
            # obtain the vertices
            for j in range(self._dag.nb_tasks_per_level[i]):
                task = self._dag.levels[i][j]
                tasks.append((task.tag,
                              {TASK_SIZE_LABEL: int(task.cost), 'alpha': round(task.alpha, 2)}))

                # Get the edges
                for k in range(task.nb_children):
                    edges.append((task.tag, task.children[k].tag, {DATA_COMM_LABEL: int(task.comm_costs[k])}))
        return tasks, edges